from typing import List

from asana import StoriesApi

from my_asana.authenticate import get_stories_api, get_tasks_api
from my_asana.users import Users


def tag_user_in_comment(user_id: str) -> str:
    return f'<a data-asana-gid="{user_id}"/>'


def add_comment_to_task(task_id: str, comment: str):
    stories_api: StoriesApi = get_stories_api()

    body = {
        "data": {
            "html_text": comment
        }
    }
    opts = {
        'opt_fields': "html_text"
    }

    stories_api.create_story_for_task(
        task_gid=task_id,
        body=body,
        opts=opts
    )


def add_followers_to_task(task_id: str):
    body = {
        "data": {
            "followers": [user.user_id for user in [Users.ORI, Users.ITAMAR, Users.HILA]]
        }
    }
    opts = {}

    get_tasks_api().add_followers_for_task(body, task_id, opts)


def get_up_for_sale_again_message() -> str:
    users_str: str = "".join([tag_user_in_comment(user.user_id) for user in [Users.ORI, Users.ITAMAR, Users.HILA]])
    main_message: str = 'This property failed in contingency. It is now back on the market! Please review it.'
    html_text: str = f'<body>{users_str}\n{main_message}</body>'
    return html_text


def get_property_sold_message() -> str:
    users_str: str = "".join([tag_user_in_comment(user.user_id) for user in [Users.ORI, Users.ITAMAR, Users.HILA]])
    main_message: str = 'This property is now sold. Moving it to Decline.'
    html_text: str = f'<body>{users_str}\n{main_message}</body>'
    return html_text

def get_off_market_message() -> str:
    main_message: str = 'This property is now off-market. Moving it to Decline.'
    html_text: str = f'<body>{main_message}</body>'
    return html_text


def log(msg: str, regular_log: bool = True, app_log: bool = False):
    """
    Log messages to console and optionally to Streamlit app.
    :param msg: Message to log
    :param regular_log: If True, log to console
    :param app_log: If True, log to Streamlit app
    """
    if regular_log:
        print(msg)
    if app_log:
        import streamlit as st
        st.write(msg)



GOOGLE_DRIVE_PROJECT_NAME: str = 'IHopesPropertiesGProject'

CLIENT_ID: str = '1040077980830-j5tkof0du3okp54iq8as0og4s5k263ja.apps.googleusercontent.com'
WEB_APP_CLIENT_ID: str = '1040077980830-hs7lpar7pel7tgcgcc8bc7tsrnm03l68.apps.googleusercontent.com'
WEB_APP_CLIENT_SECRET: str = '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' # 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

ARV_TEMPLATE_FILE_ID: str = '1etDxXHDevTsYqj6_s3T4B-CJYEoGtXh8pTf2xqu3kRE'



from typing import List

from flows.leads_generator_flow.leads_filters.abstract_filter import LeadsFilter
from property import ForSaleProperty


BEECHVIEW = ["15216"]
BROOKLINE = ["15226"]
MOUNT_WASHINGTON = ["15211"]
BRIGHTON_HEIGHTS = ["15212"]
GREENFIELD = ["15217"]
BLOOMFIELD = ["15224"]
SQUIRREL_HILL = ["15217"]
CASTLE_SHANNON = ["15234"]
DORMONT = ["15216"]
BRENWOOD = ["15227"]
MT_LEBANON = ["15228"]
MT_WASHINGTON = ["15211"]
GREEN_TREE = ["15242", "15220", "15205"]
SWISSVALE = ["15218", "15120"]


class InCityNeighborhoodsFilter(LeadsFilter):
    """
    White list of inside Pittsburgh neighborhoods
    """

    def __init__(self, in_city_zip_codes: List[List[int]]):
        super().__init__('In City Neighborhoods Filter')
        # Flatten the list of lists
        self.in_city_zip_codes: List[int] = [zip_code for zip_codes in in_city_zip_codes for zip_code in zip_codes]

    def apply_filter(self, leads: List[ForSaleProperty]) -> List[ForSaleProperty]:
        """
        Property should pass this filter if it's in the city of Pittsburgh in one of the specified zip codes or if it's
        just outside the city limits (already filtered the desired outside-Pittsburgh neighborhoods)
        :param leads:
        :return:
        """
        def is_in_city(lead: ForSaleProperty) -> bool:
            """
            If the property is in the city of Pittsburgh and in one of the specified zip codes,
             or - if it's just outside the city limits.
            :param lead:
            :return:
            """
            in_pittsburgh: bool = lead.address.city == 'Pittsburgh'
            in_city_zip_code: bool = lead.address.zip_code in self.in_city_zip_codes
            return (in_pittsburgh and in_city_zip_code) or not in_pittsburgh

        in_city_leads: List[ForSaleProperty] = [lead for lead in leads if is_in_city(lead)]

        return in_city_leads


IN_CITY_NEIGHBORHOODS_FILTER: InCityNeighborhoodsFilter = InCityNeighborhoodsFilter(
    in_city_zip_codes=[
        BEECHVIEW,
        BROOKLINE,
        MOUNT_WASHINGTON,
        BRIGHTON_HEIGHTS,
        GREENFIELD,
        BLOOMFIELD,
        SQUIRREL_HILL,
        CASTLE_SHANNON,
        DORMONT,
        BRENWOOD,
        MT_LEBANON,
        MT_WASHINGTON,
        GREEN_TREE,
        SWISSVALE
    ]
)



import math

from comps_extractor.utils.coordinates import BoundingBox


def calculate_bounding_box(lat: float, lon: float, radius: float = 0.5) -> BoundingBox:
    """
    Calculate the bounding box for a given latitude, longitude, and radius.

    :param lat: Latitude of the center point (degrees)
    :param lon: Longitude of the center point (degrees)
    :param radius: Radius around the center point (miles)
    :return: Tuple (min_lat, min_lon, max_lat, max_lon)
    """
    # Approximate degree changes per mile
    lat_change = radius / 69
    lon_change = radius / (69 * math.cos(math.radians(lat)))

    # Calculate bounding box
    min_lat = lat - lat_change
    max_lat = lat + lat_change
    min_lon = lon - lon_change
    max_lon = lon + lon_change

    print(f"Created Bounding box for {lat}, {lon} with radius {radius} miles")
    print(f"Bounding box: min_lat={min_lat}, min_lon={min_lon}, max_lat={max_lat}, max_lon={max_lon}")

    return BoundingBox(
        max_lat=max_lat,
        max_lon=max_lon,
        min_lat=min_lat,
        min_lon=min_lon
    )



from typing import List

from apify.zillow_scrapers.zillow_detail_scraper.on_the_fly_runner import generate_for_sale_property_from_address
from common.utils.logger import log
from comps_extractor.comps_generator import generate_comps
from my_asana.tasks_generator import generate_new_property_task
from my_asana.utils import is_task_exists
from property import ForSaleProperty, SoldProperty

def address_based_lead_flow(address: str) -> str:
    lead_property: ForSaleProperty = generate_for_sale_property_from_address(address=address)
    property_link: str = lead_flow(lead_property)
    return property_link


def lead_flow(lead_property: ForSaleProperty) -> str:
    address: str = lead_property.address.get_full_address()
    log(msg=f'Working on property: {address}')

    if is_task_exists(address):
        log(msg=f"Task already exists for: {address}, stopping the process..")
        return ''

    lead_property: ForSaleProperty = generate_for_sale_property_from_address(address=address)
    log(msg=f"Generated a new for sale property successfully: {lead_property}")

    comps: List[SoldProperty] = generate_comps(
        for_sale_property=lead_property,
        test_mode=False
    )
    address: str = lead_property.address.get_full_address()
    log(f'Found total of {len(comps)} comps for {address}')

    if not comps:  # If no comps found, add to the no-comps list
        log(msg=f"No comps found for property: {address}.")

    property_link: str = generate_new_property_task(
        for_sale_property=lead_property,
        comps=comps
    )
    log(msg=f'Generated Asana task for {address} with link: {property_link}')

    return property_link



import pickle
import os
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.discovery import build, Resource

SCOPES = [
    'https://www.googleapis.com/auth/gmail.send',
    'https://www.googleapis.com/auth/gmail.labels',
    'https://www.googleapis.com/auth/gmail.modify'
]

def authenticate() -> Resource:
    creds = None
    # Check if token.pickle exists
    if os.path.exists('token.pickle'):
        with open('token.pickle', 'rb') as token:
            creds = pickle.load(token)

    # If credentials don't exist or are invalid, get new ones
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                'credentials.json', SCOPES)
            creds = flow.run_local_server(port=0)

        # Save the credentials for the next run
        with open('token.pickle', 'wb') as token:
            pickle.dump(creds, token)

    # Build and return the service
    service = build('gmail', 'v1', credentials=creds)
    return service


from urllib.parse import quote


def construct_zillow_basic_property_url(property_address: str) -> str:
    """
    Constructs a Zillow property URL by encoding the address to be URL-safe.
    :param property_address: The address of the property, e.g., "123 Main St, Anytown, USA"
    :return: A URL string pointing to the Zillow listing
    """
    encoded_address = quote(property_address)
    return f'https://www.zillow.com/homes/{encoded_address}'




class AddressCoordinates:
    def __init__(self, address: str, lat: float, lon: float):
        self.address = address
        self.lat = lat
        self.long = lon

class BoundingBox:
    def __init__(self, max_lat: float, max_lon: float, min_lat: float, min_lon: float):
        self.max_lat = max_lat
        self.max_lon = max_lon
        self.min_lat = min_lat
        self.min_lon = min_lon



import json
from enum import Enum
from urllib.parse import quote

from comps_extractor.utils.coordinates import BoundingBox
from property import Property


class ListingStatus(Enum):
    SOLD = "sold"
    FOR_SALE = "for_sale"
    FOR_RENT = "for_rent"


def generate_bonding_box_based_zillow_url(for_sale_property: Property, bounding_box: BoundingBox,
                                          sold_in_last_months: int = 6, status: ListingStatus = ListingStatus.SOLD)\
        -> str:
    """
    Generate a Zillow search URL dynamically with filters for sold properties.
    :param status:
    :param for_sale_property:
    :param state: State abbreviation (e.g., "pa").
    :param bounding_box: BoundingBox object containing min/max lat/lon.
    :param region_id: Zillow region ID.
    :param zoom: Map zoom level (default: 17).
    :param sold_in_last_months: Filter for properties sold within the last X months (default: 6).
    :return: Zillow search URL as a string.
    """
    # Base URL
    base_url = (f"https://www.zillow.com/"
                f"{quote(for_sale_property.address.city)}-{for_sale_property.address.state}/{status.value}/")

    # Filter for sold properties in the last X months
    date_range_filter = {
        1: "30",
        3: "90",
        6: "6m",
        12: "12m",
        24: "24m",
    }.get(sold_in_last_months, "6m")  # Default to "6m"

    # Build filterState
    filter_state = {
        "sort": {
            "value": "globalrelevanceex"
        },
        "fsba": {
            "value": False
        },
        "fsbo": {
            "value": False
        },
        "nc": {
            "value": False
        },
        "cmsn": {
            "value": False
        },
        "auc": {
            "value": False
        },
        "fore": {
            "value": False
        },
        "rs": {
            "value": True
        },
        "doz": {
            "value": date_range_filter
        },
        "beds": {
            "min": for_sale_property.bedrooms,
            "max": for_sale_property.bedrooms
        },
    }

    if for_sale_property.sqft:
        filter_state["sqft"] = {
            "min": for_sale_property.get_min_sqft(),
            "max": for_sale_property.get_max_sqft()
        }

    # Add property related filters

    # Build the searchQueryState JSON object
    search_query_state = {
        "pagination": {},
        "isMapVisible": True,
        "mapBounds": {
            "west": bounding_box.min_lon,
            "east": bounding_box.max_lon,
            "south": bounding_box.min_lat,
            "north": bounding_box.max_lat,
        },
        # "regionSelection": [
        #     {
        #         "regionId": for_sale_property.address.zip_code,
        #         "regionType": 7
        #     }
        # ],
        "filterState": filter_state,
        "isListVisible": True,
        "usersSearchTerm": f"{for_sale_property.address.city} {for_sale_property.address.state}".title(),
    }

    # Encode the JSON object as a URL-safe string
    search_query_state_encoded = quote(json.dumps(search_query_state))

    # Construct the full URL
    full_url = f"{base_url}?searchQueryState={search_query_state_encoded}"
    print(f'For the property {for_sale_property.address.get_full_address()}, the Zillow URL for its comps is: {full_url}')

    return full_url



import json
from typing import List

# File to store properties with no comps
NO_COMPS_FILE = 'src/leads_generator_flow/no_comps_handling/no_comps_properties.json'

# Load the list of properties with no comps from the file
def load_no_comps_properties():
    try:
        with open(NO_COMPS_FILE, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        return []  # Return an empty list if the file doesn't exist

# Save the list of properties with no comps to the file
def save_no_comps_properties(no_comps_list):
    with open(NO_COMPS_FILE, "w") as f:
        json.dump(no_comps_list, f)

# Check if a property is in the no-comps list
def is_in_no_comps_list(property_address):
    no_comps_list = load_no_comps_properties()
    return property_address in no_comps_list

# Add a property to the no-comps list
def add_to_no_comps_list(property_address):
    no_comps_list = load_no_comps_properties()
    if property_address not in no_comps_list:
        no_comps_list.append(property_address)
        save_no_comps_properties(no_comps_list)


import copy

from common.utils.logger import log
from property import PropertyAddress


def update_address_with_pittsburgh_ciy(address: PropertyAddress) -> PropertyAddress:
    """
    We might have cases where the property's address is written with a city such as Baden or Penn Hills in some places
     while in others it is listed with Pittsburgh as city.
    This usually can happen as the task is originally created using Rentcast API which will fetch the address as
     city=Baldwin, although in Zillow it will appear with City=Pittsburgh.
     Also, it might appear with city=Baldwin even in Redfin.
    :param address: Original address
    :return: Updated address with city set to Pittsburgh
    """
    log(msg=f"Got address: {address.get_full_address()}", app_log=False)
    address_copy: PropertyAddress = copy.deepcopy(address)
    address_copy.city = 'Pittsburgh'
    log(msg=f"Updated address: {address_copy.get_full_address()}", app_log=False)
    return address_copy



import streamlit as st

from flows.leads_cleaner_flow.flow import run_leads_cleaner


def run_leads_cleaner_flow():
    st.title("🧹 Leads Cleaner 🧹")
    st.write("This flow will help clean and remove already sold/under contract properties from the leads section.")

    if st.button("Clean Leads"):
        with st.spinner("Running..."):
            try:
                st.info(f"Leads cleaner flow has been triggered.")
                run_leads_cleaner()
                st.success(f"Leads cleaner flow has been completed.")
            except Exception as e:
                st.error(f"Error: {e}")



import asana
from asana import TasksApi, ApiClient, StoriesApi

from my_asana.consts import PERSONAL_ACCESS_TOKEN


def get_tasks_api() -> TasksApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    tasks_api = asana.TasksApi(api_client)
    return tasks_api


def get_stories_api() -> StoriesApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    stories_api = asana.StoriesApi(api_client)
    return stories_api


def get_project_api() -> asana.ProjectsApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    projects_api = asana.ProjectsApi(api_client)
    return projects_api


def get_custom_fields_api() -> asana.CustomFieldsApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    custom_fields_api = asana.CustomFieldsApi(api_client)
    return custom_fields_api


def get_workspace_api() -> asana.WorkspacesApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    workspace_api = asana.WorkspacesApi(api_client)
    return workspace_api


def get_tags_api() -> asana.TagsApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    tags_api = asana.TagsApi(api_client)
    return tags_api


def get_users_api() -> asana.UsersApi:
    # Configure access
    api_client = get_api_client()

    # Create API instance
    users_api = asana.UsersApi(api_client)
    return users_api


def get_api_client() -> ApiClient:
    configuration = asana.Configuration()
    configuration.access_token = PERSONAL_ACCESS_TOKEN
    api_client = asana.ApiClient(configuration)
    return api_client



import os
import sys
from datetime import datetime
import streamlit as st

sys.path.append(os.path.abspath("src"))

from flows.leads_generator_flow.flow.lead_flow import address_based_lead_flow
from google_drive.authenticator import get_google_services, GoogleServices

LEADS = []

def create_lead(address):
    return {
        "id": len(LEADS) + 1,
        "address": address,
        "status": "processing",
        "processedDate": datetime.now().isoformat()
    }

def update_lead(lead, **kwargs):
    lead.update(kwargs)

def run_single_lead_analyzer_flow():
    st.title("🏡 Property Lead Analyzer 🏡")
    st.write("Enter a property address to retrieve its comps and create Asana task.")

    if "current_lead" not in st.session_state:
        st.session_state.current_lead = None
    if "error" not in st.session_state:
        st.session_state.error = None
    if 'current_address' not in st.session_state:
        st.session_state.current_address = None

    address = st.text_input("Property Address")
    if st.button("Process"):
        if address:
            st.session_state.error = None
            st.session_state.current_address = address
            lead = create_lead(address)
            LEADS.insert(0, lead)
            st.session_state.current_lead = lead

            google_services: GoogleServices = get_google_services()
            if google_services is None:
                st.warning("Authentication failed or not completed.")
                st.stop()

            with st.spinner("Processing property..."):
                try:
                    property_link: str = address_based_lead_flow(address=address)
                    update_lead(lead, status="completed", property_link=property_link)
                except Exception as e:
                    update_lead(lead, status="failed")
                    st.session_state.error = str(e)
        else:
            st.warning("Please enter an address.")

    if st.session_state.error:
        st.error(f"Error: {st.session_state.error}")

    if st.session_state.current_lead and st.session_state.current_lead["status"] == "completed":
        lead = st.session_state.current_lead
        if 'property_link' in lead:
            st.success("Processing complete! See link to Asana task below.")
        else:
            st.warning(f"Couldn't create a new task")
        st.json(lead)

    st.subheader("📜 Recent Leads")
    if LEADS:
        st.table([{
            "Address": l["address"],
            "Status": l["status"],
            "Date": l["processedDate"][:10],
            "Asana task": l['property_link'] if 'property_link' in l else "N/A",
        } for l in LEADS])
    else:
        st.info("No leads yet.")



from authenticate import authenticate
from typing import List

class LabelManager:
    def __init__(self, service = authenticate()):
        """
        Initialize the LabelManager instance.
        """
        self.service = service

    def add_label_to_email(self, message_id: str, label_names: List[str]):
        """
        Adds labels to an email and returns True if successful, otherwise False.
        :param message_id: ID of the email.
        :param label_names: List of label names to add.
        """
        try:
            # Step 1: Get the list of labels
            labels = self._get_labels()
            label_ids = []

            # Step 2: Find or create the label
            for label_name in label_names:
                label_id = None
                for label in labels:
                    if label["name"] == label_name:
                        label_id = label["id"]
                        break

                if not label_id:
                    new_label = self._create_label(label_name)
                    label_id = new_label["id"]

                label_ids.append(label_id)

            # Step 3: Add the label to the email
            self.service.users().messages().modify(
                userId="me",
                id=message_id,
                body={"addLabelIds": label_ids}
            ).execute()
            print(f"Labels {', '.join(label_names)} added successfully to email")
            return True
        except Exception as e:
            print(f"Error adding label: {e}")
            return False

    def _get_labels(self):
        """
        Retrieves all labels in the user's Gmail account.
        :return: List of labels.
        """
        labels = self.service.users().labels().list(userId="me").execute().get("labels", [])
        return labels

    def _create_label(self, label_name):
        """
        Creates a new label.
        :param label_name: Name of the label.
        :return: Created label details.
        """
        label_body = {
            "name": label_name,
            "labelListVisibility": "labelShow",
            "messageListVisibility": "show"
        }
        return self.service.users().labels().create(userId="me", body=label_body).execute()



from enum import Enum


def get_new_property_media(property_address: str):
    return ""


class TaskPlaceHolder(Enum):
    ARV_DOC = "ARV Doc"
    PROPERTY_LINK = "Property Link"
    PROPERTY_MEDIA = "Property Media"


# PLACE_HOLDERS_APPLIERS: Dict[TaskPlaceHolder, Callable[[str], str]] = {
#     TaskPlaceHolder.ARV_DOC: save_as_template_arv,
#     TaskPlaceHolder.PROPERTY_LINK: construct_zillow_property_url,
#     TaskPlaceHolder.PROPERTY_MEDIA: get_new_property_media
# }



from typing import List, Optional

from comps_extractor.utils.bounding_box_calculator import calculate_bounding_box
from comps_extractor.utils.coordinates import AddressCoordinates, BoundingBox
from comps_extractor.utils.geo_encoder import get_coordinates
from apify.zillow_scrapers.zillow_search_scraper.scraper import scrape_zillow
from zillow.zillow_url_generator import generate_bonding_box_based_zillow_url
from property import Property, SoldProperty, ForSaleProperty, PropertyAddress


def generate_comps(for_sale_property: ForSaleProperty, radius: Optional[float] = 0.5,
                   sold_in_last_months: Optional[int] = 6, test_mode: bool = False) -> List[SoldProperty]:
    """
    Generate comps for a for-sale property.
    :param for_sale_property: The for-sale property for which to find comps.
    :param radius: Radius in miles to search for comps.
    :param sold_in_last_months: Sold properties in the last N months.
    :param test_mode: Test mode flag.
    :return:
    """

    # Step 1: Get coordinates for the address
    if not for_sale_property.lat or not for_sale_property.long:
        address_coordinates: Optional[AddressCoordinates] = get_coordinates(for_sale_property.address)
        if not address_coordinates:
            print(f"Failed to get coordinates for address: {for_sale_property.address.get_full_address()}")
            return []
        for_sale_property.lat = address_coordinates.lat
        for_sale_property.long = address_coordinates.long
        print(f"Got coordinates successfully.")

    # Step 2: Calculate bounding box
    bounding_box: BoundingBox = calculate_bounding_box(
        lat=for_sale_property.lat,
        lon=for_sale_property.long,
        radius=radius
    )
    print(f"Created Bounding box successfully.")

    # Step 3: Construct Zillow URL with bounding box
    zillow_url = generate_bonding_box_based_zillow_url(
        for_sale_property=for_sale_property,
        bounding_box=bounding_box,
        sold_in_last_months=sold_in_last_months
    )
    print("Generated Zillow URL:", zillow_url)

    # Step 4: Scrape Zillow for comps
    scraped_properties: List[SoldProperty] = scrape_zillow(
        lead_property=for_sale_property,
        zillow_url=zillow_url,
        test_mode=test_mode
    )

    # We might have a scenario where the for-sale property is also in the comps list (if sold in the last 6 months),
    # hence removing it
    final_scraped_properties: List[SoldProperty] = \
        [prop for prop in scraped_properties
         if prop.address.get_full_address() != for_sale_property.address.get_full_address()]

    return final_scraped_properties



import base64
from email.mime.text import MIMEText
from typing import List, Optional
from authenticate import authenticate
from label_manager import LabelManager

# Import settings from consts file
from consts import EMAIL_ADDRESS

class EmailSender:
    def __init__(self):
        """
        Initialize the EmailSender instance.
        """
        self.email_address: str = EMAIL_ADDRESS

    def send_email(self, subject: str, body: str, recipient: List[str] = [EMAIL_ADDRESS], cc: Optional[List[str]] = None, bcc: Optional[List[str]] = None, label_names: Optional[List[str]] = None):
        """
        Sends an email using the Gmail API.
        :param recipient: Recipient email address.
        :param subject: Email subject.
        :param body: Email body content.
        :param cc: The CC recipient of the email.
        :param bcc: The BCC recipient of the email.
        :param label_names: Name of the labels to add to the email message.
        """
        raw_msg = self._create_message(subject, body, recipient, cc, bcc)
        service = authenticate()
        sent_message: dict[str, List[str]] = service.users().messages().send(userId="me", body=raw_msg).execute()
        print(f"Email sent successfully. Message ID: {sent_message['id']}")
        if label_names:
            LabelManager(service).add_label_to_email(sent_message['id'], label_names)

    def _create_message(self, subject: str, body: str, recipient: List[str] = EMAIL_ADDRESS, cc: Optional[List[str]] = None, bcc: Optional[List[str]] = None) -> dict:
        """
        Creates an email message.
        :param recipient: Recipient email address.
        :param subject: Email subject.
        :param body: Email body content.
        :param cc: The CC recipient of the email.
        :param bcc: The BCC recipient of the email.
        :return: Encoded email message.
        """

        msg = MIMEText(body)
        msg['From'] = self.email_address
        msg['To'] = self.handel_email_addresses(recipient)
        if cc:
            msg['Cc'] = self.handel_email_addresses(cc)
        if bcc:
            msg['Bcc'] = self.handel_email_addresses(bcc)
        msg['Subject'] = subject
        return {'raw': base64.urlsafe_b64encode(msg.as_bytes()).decode()}


    @staticmethod
    def handel_email_addresses(addresses: List[str]) -> str:
        """
        Return a string of email addresses separated by commas.
        :param addresses: A list of email addresses.
        :return: A string of email addresses separated by commas.
        """
        return ", ".join(addresses)



from datetime import date
from typing import List, Type

from asana import TasksApi

from apify.zillow_scrapers.zillow_detail_scraper.contingency_status_extractor import get_contingency_status
from flows.under_contract_follow_up_flow.contingent_status_handlers import ContingentStatus, ContingentStatusMapper, \
    AbstractContingentHandler
from flows.under_contract_follow_up_flow.get_under_contract_properties import get_under_contract_properties
from flows.under_contract_follow_up_flow.under_contract_task import UnderContractTask
from my_asana.authenticate import get_tasks_api


def run_under_contract_leads_flow():
    """
    Run the under contract follow-up flow, will update the under contract properties according to their status.
    :return:
    """
    print('Running under contract follow-up flow')
    under_contract_properties: List[UnderContractTask] = get_under_contract_properties()
    print(f'Found {len(under_contract_properties)} under contract properties')

    tasks_api: TasksApi = get_tasks_api()
    for under_contract_property in under_contract_properties:
        property_address: str = under_contract_property.address.get_full_address()
        print(f'Under contract property: {property_address}')

        status: ContingentStatus = get_contingency_status(under_contract_task=under_contract_property)
        print(f'For property {property_address}, status is {status.value}')

        # Update the due date to today
        today: str = date.today().strftime("%Y-%m-%d")
        tasks_api.update_task(
            task_gid=under_contract_property.gid,
            body={"data": {"due_on": today}},
            opts={}
        )

        handler_class: Type[AbstractContingentHandler] = ContingentStatusMapper().get_handler(status)
        handler_instance: AbstractContingentHandler = handler_class(under_contract_property)
        handler_instance.handle()
        print(f'Finished handling property: {property_address}')


if __name__ == '__main__':
    run_under_contract_leads_flow()



from apify.zillow_scrapers.zillow_detail_scraper.scraper import scrape_zillow_with_retry
from flows.under_contract_follow_up_flow.contingent_status_handlers import ContingentStatus
from flows.under_contract_follow_up_flow.under_contract_task import UnderContractTask
from property import PropertyAddress


def get_contingency_status(under_contract_task: UnderContractTask) -> ContingentStatus:
    """
    Run the Apify Zillow scraper using the Apify client.
    :param under_contract_task: The under contract property to scrape.
    :return: Details of the scraper execution, including a link to the dataset.
    """

    return get_contingency_status_by_address(under_contract_task.address)


def get_contingency_status_by_address(property_address: PropertyAddress) -> ContingentStatus:
    """
    Run the Apify Zillow scraper using the Apify client.
    :param property_address: The address of the property to scrape.
    :return: Details of the scraper execution, including a link to the dataset.
    """

    property_data: dict = scrape_zillow_with_retry(property_address)
    contingency_status: ContingentStatus = transform_property_data_to_contingency_status(property_data)
    return contingency_status


def transform_property_data_to_contingency_status(property_data: dict) -> ContingentStatus:
    """
    Transform the response data from the Zillow scraper into a ContingentStatus object.
    :param property_data:
    :return:
    """
    still_under_contract: bool = (property_data['contingentListingType'] == 'UNDER_CONTRACT' or
                                  property_data['homeStatus'] == 'PENDING')

    is_for_sale: bool = property_data['homeStatus'] == 'FOR_SALE'

    is_sold: bool = property_data['homeStatus'] == 'RECENTLY_SOLD' or property_data['homeStatus'] == 'SOLD' or \
                    property_data['homeStatus'] == 'FORECLOSED'

    is_off_market: bool = property_data['homeStatus'] == 'OTHER' and property_data['adTargets']['listtp'] == 'not_for_sale'

    if not any([still_under_contract, is_for_sale, is_sold, is_off_market]):
        print(f"Property data: {property_data}")
        raise ValueError(f"Could not determine the status of the property: {property_data}")

    if still_under_contract:
        return ContingentStatus.UNDER_CONTRACT
    elif is_for_sale:
        return ContingentStatus.FOR_SALE
    elif is_sold:
        return ContingentStatus.RECENTLY_SOLD
    elif is_off_market:
        return ContingentStatus.OFF_MARKET
    else:
        raise ValueError(f"Could not determine the status of the property: {property_data}")



import asana
from asana.rest import ApiException
from pprint import pprint

from my_asana.authenticate import get_api_client
from my_asana.consts import LEADS_PROJECT_ID

def get_sections_data():
    api_client = get_api_client()

    # create an instance of the API class
    sections_api_instance = asana.SectionsApi(api_client)

    opts = {
        'limit': 50, # int | Results per page. The number of objects to return per page. The value must be between 1 and 100.
        'opt_fields': "created_at,name,offset,path,project,project.name,projects,projects.name,uri", # list[str] | This endpoint returns a compact resource, which excludes some properties by default. To include those optional properties, set this query parameter to a comma-separated list of the properties you wish to include.
    }

    try:
        # Get sections in a project
        api_response = sections_api_instance.get_sections_for_project(LEADS_PROJECT_ID, opts)
        for data in api_response:
            pprint(data)
    except ApiException as e:
        print("Exception when calling SectionsApi->get_sections_for_project: %s\n" % e)


if __name__ == "__main__":
    get_sections_data()


from abc import ABC, abstractmethod
from typing import List

from property import ForSaleProperty


class LeadsFilter(ABC):

    def __init__(self, name: str):
        self.name: str = name

    @abstractmethod
    def apply_filter(self, leads: List[ForSaleProperty]) -> List[ForSaleProperty]:
        pass



from datetime import date
from typing import List, Dict

from asana import TasksApi

from flows.under_contract_follow_up_flow.under_contract_task import UnderContractTask
from my_asana.authenticate import get_tasks_api
from my_asana.consts import FOLLOW_UPS_UNDER_CONTRACT_SECTION


def get_today_str() -> str:
    return date.today().strftime("%Y-%m-%d")


def get_task_due_date(tasks_api: TasksApi, under_contract_task: UnderContractTask) -> str:
    return tasks_api.get_task(under_contract_task.gid, opts={})['due_on']


def get_under_contract_properties() -> List[UnderContractTask]:
    tasks_api: TasksApi = get_tasks_api()
    raw_tasks: List[Dict[str, str]] = list(tasks_api.get_tasks_for_section(
        section_gid=FOLLOW_UPS_UNDER_CONTRACT_SECTION,
        async_req=False,
        opts={})
    )

    under_contract_tasks: List[UnderContractTask] = \
        [UnderContractTask(address=raw_task['name'], gid=raw_task['gid']) for raw_task in raw_tasks]
    print(f'Found {len(under_contract_tasks)} under contract properties')

    today_str: str = get_today_str()

    # Filter only properties that we didn't analyze yet, i.e. the due date is not today (not set already)
    filtered_under_contract_tasks: List[UnderContractTask] = [
        under_contract_task for under_contract_task in under_contract_tasks
        if get_task_due_date(tasks_api, under_contract_task) != today_str
    ]
    print(f'Remain {len(filtered_under_contract_tasks)} under contract properties (not analyzed yet)')

    return filtered_under_contract_tasks



from typing import List

from asana.rest import ApiException

from google_drive.duplicate_template_arv import save_as_template_arv
from my_asana.authenticate import get_tasks_api
from my_asana.consts import PROPERTY_TASK_TEMPLATE_ID, LEADS_PROPERTIES_SECTION, LEADS_PROJECT_ID
from my_asana.task_place_holders import TaskPlaceHolder, get_new_property_media
from my_asana.utils import add_task_to_section
from property import SoldProperty, ForSaleProperty
from zillow.property_getter import construct_zillow_basic_property_url


def generate_new_property_task(for_sale_property: ForSaleProperty, comps: List[SoldProperty]) -> str:
    # Create API instance
    property_address: str = for_sale_property.address.get_full_address()

    try:
        # Get the source task
        tasks_api = get_tasks_api()
        template_task = tasks_api.get_task(PROPERTY_TASK_TEMPLATE_ID, opts=[])
        template_description = template_task['notes']

        for place_holder in TaskPlaceHolder:
            if place_holder.value in template_description:
                print(f"Found placeholder: {place_holder.value}")
                if place_holder is TaskPlaceHolder.ARV_DOC:
                    updated_description: str = save_as_template_arv(for_sale_property, comps)
                elif place_holder is TaskPlaceHolder.PROPERTY_LINK:
                    updated_description: str = construct_zillow_basic_property_url(property_address)
                elif place_holder is TaskPlaceHolder.PROPERTY_MEDIA:
                    updated_description: str = get_new_property_media(property_address)
                else:
                    raise ValueError(f"Unknown placeholder: {place_holder.value}")

                template_description = template_description.replace(
                    f'{place_holder.value}:',
                    f"{place_holder.value}: {updated_description}"
                )
            else:
                print(f"Placeholder not found: {place_holder.value}. Having an issue here.")

        # Create a new task
        new_task_body = {
            "data": {
                "name": property_address,
                "notes": template_description,
                "projects": LEADS_PROJECT_ID,  # Copy the same project association
            }
        }

        new_task = tasks_api.create_task(new_task_body, opts=[])
        add_task_to_section(task_id=new_task["gid"], section_id=LEADS_PROPERTIES_SECTION)
        print("New task created successfully!")
        print("New Task ID:", new_task['gid'])
        #print("New Task Description:", new_task['notes'])
        task_link: str = f"https://app.asana.com/0/{LEADS_PROJECT_ID}/{new_task['gid']}/f"
        print(f'Task link for {property_address}: {task_link}')
        return task_link

    except ApiException as e:
        print("Error occurred:", e)



from typing import List

from flows.leads_generator_flow.flow.lead_flow import address_based_lead_flow
from google_drive.authenticator import get_google_services

if __name__ == '__main__':

    addresses: List[str] = [
        '211 Ohio St, Monroeville, PA 15146',
    ]

    print('Authenticating...')
    get_google_services()
    print('Authenticated successfully.')

    for address in addresses:
        print(f'Working on address: {address}')
        address_based_lead_flow(address=address)



import requests
from my_asana.consts import PERSONAL_ACCESS_TOKEN
from enum import Enum

def get_users():
    url = "https://app.asana.com/api/1.0/users"
    headers = {
        "Authorization": f"Bearer {PERSONAL_ACCESS_TOKEN}"
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return response.json()['data']
    else:
        print(f"Error fetching users: {response.status_code} - {response.text}")
        return []

#users = get_users()
#print(users)  # This will give you a list of users and their IDs

class Users(Enum):
    ITAMAR = 'Itamar Halevi', '1206867055887039'
    ORI = 'Ori Peri', '1206866397949967'
    HILA = 'Hila Lichtman Peri', '1207911884233138'
    U25TO40 = '25 to 40', '1207911699123552'

    def __init__(self, user_name: str, user_id: str) -> None:
        self.user_name: str = user_name
        self.user_id: str = user_id 


def randomly_select_user_id() -> str:
    import random
    # Create a list of user IDs excluding U25TO40
    user_ids = [user.user_id for user in Users if user in [Users.ITAMAR, Users.ORI]]

    # Select a random user ID from the filtered list
    selected_user_id = random.choice(user_ids)

    return selected_user_id



from typing import List, Dict

from asana import TasksApi

from flows.leads_cleaner_flow.leads_task import LeadsTask
from my_asana.authenticate import get_tasks_api
from my_asana.consts import LEADS_PROPERTIES_SECTION, PROPERTY_TASK_TEMPLATE_ID


def get_leads_properties() -> List[LeadsTask]:
    tasks_api: TasksApi = get_tasks_api()
    raw_tasks: List[Dict[str, str]] = list(tasks_api.get_tasks_for_section(
        section_gid=LEADS_PROPERTIES_SECTION,
        async_req=False,
        opts={})
    )

    leads_tasks: List[LeadsTask] = \
        [LeadsTask(address=raw_task['name'], gid=raw_task['gid']) for raw_task in raw_tasks \
         if raw_task['gid'] != PROPERTY_TASK_TEMPLATE_ID]
    print(f'Found {len(leads_tasks)} leads properties')

    return leads_tasks



from typing import List, Optional

from apify.zillow_scrapers.common.client import MyApifyClient
from apify.zillow_scrapers.common.config import Actor
from property import PropertyAddress, SoldProperty, ForSaleProperty, Property


def scrape_zillow(lead_property: ForSaleProperty, zillow_url: str, test_mode: bool = False) -> List[SoldProperty]:
    """
    Run the Apify Zillow scraper using the Apify client.

    :param lead_property: The for-sale property for which to find comps.
    :param zillow_url: The Zillow search URL to scrape.
    :return: Details of the scraper execution, including a link to the dataset.
    """
    # Initialize the Apify client
    apify_client: MyApifyClient = MyApifyClient(actor=Actor.ZILLOW_SEARCH_SCRAPER)

    # Prepare the Actor input
    run_input = {
        "searchUrls": [
            {
                "url": zillow_url
            }
        ],
    }

    # Fetch and print Actor results from the run's dataset (if there are any)
    scraped_properties: List[SoldProperty] = []
    raw_scraped_properties: Optional[List[dict]] = apify_client.run_client(run_input=run_input)
    if raw_scraped_properties:
        for item in raw_scraped_properties:
            print(item)
            scraped_properties.append(transform_response_to_property(
                lead_property=lead_property,
                response_data_dict=item
            ))

    return scraped_properties


def transform_response_to_property(lead_property: ForSaleProperty, response_data_dict: dict) -> SoldProperty:
    """
    Transform the response data from the Zillow scraper into a list of Property objects.

    :param response_data_dict: The response data from the Zillow scraper.
    :return: A list of Property objects.
    """

    return SoldProperty(
        address=PropertyAddress(
            address=response_data_dict.get("addressStreet"),
            city=response_data_dict.get("addressCity"),
            state=response_data_dict.get("addressState"),
            zip_code=response_data_dict.get("addressZipcode")
        ),
        zillow_pid=response_data_dict.get("zpid"),
        zillow_link=response_data_dict.get("detailUrl"),
        beds=response_data_dict.get("beds"),
        baths=response_data_dict.get("baths"),
        sqft=response_data_dict.get("area"),
        lot_sqft=response_data_dict.get("hdpData").get("homeInfo").get("lotAreaValue"),  # Convert acres to sqft
        lat=response_data_dict.get("latLong").get("latitude"),
        long=response_data_dict.get("latLong").get("longitude"),
        dist_from_lead=round(Property.calculate_distance(
            lat1=lead_property.lat,
            lon1=lead_property.long,
            lat2=response_data_dict.get("latLong").get("latitude"),
            lon2=response_data_dict.get("latLong").get("longitude")
        ),2),
        sold_price=response_data_dict.get("unformattedPrice"),
        sold_date=get_sold_date(response_data_dict),
    )


def get_sold_date(response_data_dict):
    try:
        from datetime import datetime
        # Milliseconds timestamp
        millis = response_data_dict['hdpData']['homeInfo']['dateSold']  # Example: corresponds to '2023-01-01 00:00:00'
        # Convert milliseconds to seconds (datetime expects seconds)
        seconds = millis / 1000
        # Convert to a datetime object
        date = datetime.fromtimestamp(seconds)
        # Format as a readable string
        sold_date = date.strftime('%d-%m-%Y')
    except Exception as e:
        raise ValueError(f"Could not extract sold date from response data: {e}. "
                         f"We cannot extract comps without a sold date.")
    return sold_date



from typing import List

import requests

from flows.leads_generator_flow.flow.lead_source import LeadQuerySource
from property import PropertyAddress, ForSaleProperty

API_KEY = "b632a2d8212f4bb0bb88fc43dd6b976f"
BASE_URL = "https://api.rentcast.io/v1"

def fetch_active_for_sale_properties(lead_source: LeadQuerySource, test_mode: bool = False) -> List[ForSaleProperty]:
    """
    Fetch listings from RentCast API.
    """

    if test_mode:
        import pickle
        with open('for_sale_properties_test_data.pickle', 'rb') as token:
            return pickle.load(token)

    endpoint = f"{BASE_URL}/listings/sale"

    headers = {
        "accept": "application/json",
        "X-Api-Key": API_KEY
    }

    limit: int = 500
    params = {
        #"latitude": downtown_pittsburgh_latitude,
        #"longitude": downtown_pittsburgh_longitude,
        #"radius": radius,
        "city": lead_source.city,
        "state": lead_source.state,
        "propertyType": 'Single Family',
        "status": 'Active',
        "limit": limit
    }

    if lead_source.zipcode:
        params["zipCode"] = lead_source.zipcode

    # Iterate as long as there are more results
    offset = 0
    for_sale_properties: List[ForSaleProperty] = []
    while True:
        curr_offset_properties: List[ForSaleProperty] = pagination_request(endpoint, headers, params, offset)
        for_sale_properties.extend(curr_offset_properties)

        if not curr_offset_properties:
            print(f"No properties found at all. Stopping the process.")
            break

        fetched_all_properties: bool = len(curr_offset_properties) < limit
        if fetched_all_properties:
            break

        offset += limit

    sorted_for_sale_properties: List[ForSaleProperty] = sort_for_sale_properties_by_listing_date(for_sale_properties)
    return sorted_for_sale_properties


def sort_for_sale_properties_by_listing_date(for_sale_properties: List[ForSaleProperty]) -> List[ForSaleProperty]:
    """
    Sort for sale property by their listing date.
    By that - we can iterate and stop when we reach the first property that we already have a task for.
    :param for_sale_properties:
    :return:
    """
    return sorted(for_sale_properties, key=lambda x: x.listed_date)


def pagination_request(endpoint: str, headers: dict, params: dict, offset: int) -> List[ForSaleProperty]:
    """
    Make a request to the RentCast API with pagination.
    :param endpoint:
    :param headers:
    :param params:
    :param offset:
    :return:
    """

    # API has a 20 requests per second limit, sleep accordingly
    import time
    time.sleep(0.05)  # 50 ms, 1/20 = 0.05

    params["offset"] = offset
    response = requests.get(endpoint, headers=headers, params=params)
    if response.status_code == 200:
        response_dicts: List[dict] = response.json()
        for_sale_properties: List[ForSaleProperty] = \
            [transform_response_to_property(response_dict) for response_dict in response_dicts]
        return for_sale_properties
    else:
        response.raise_for_status()


def transform_response_to_property(response_data_dict: dict) -> ForSaleProperty:
    """
    Transform the response data from the RentCast API into a ForSaleProperty.

    :param response_data_dict: The response data from the RentCast API.
    :return: A ForSaleProperty object.
    """

    return ForSaleProperty(
        rentcast_pid=response_data_dict.get("id"),
        address=PropertyAddress.from_full_address(response_data_dict.get('formattedAddress')),
        beds=response_data_dict.get("bedrooms"),
        baths=response_data_dict.get("bathrooms"),
        sqft=response_data_dict.get("squareFootage"),
        lot_sqft=response_data_dict.get("lotSize"),
        lat=response_data_dict.get("latitude"),
        long=response_data_dict.get("longitude"),
        year_built=response_data_dict.get("yearBuilt"),
        asking_price = response_data_dict.get("price"),
        listing_type=response_data_dict.get("listingType"),
        listed_date=response_data_dict.get("listedDate")[:10],  # Keep only year-month-day
        days_on_market=response_data_dict.get("daysOnMarket")
    )



from apify.zillow_scrapers.zillow_detail_scraper.scraper import scrape_zillow_with_retry
from property import ForSaleProperty, PropertyAddress


def generate_for_sale_property_from_address(address: str) -> ForSaleProperty:
    """
    Generate a ForSaleProperty object from the address.
    :param address:
    :return:
    """

    property_data: dict = scrape_zillow_with_retry(PropertyAddress.from_full_address(address))
    address_dict: dict = property_data['address']
    property_address: PropertyAddress = PropertyAddress(
        address=address_dict['streetAddress'],
        city=address_dict['city'],
        state=address_dict['state'],
        zip_code=address_dict['zipcode']
    )

    return ForSaleProperty(
        address=property_address,
        beds=property_data['bedrooms'],
        baths=property_data['bathrooms'],
        lat=property_data['latitude'],
        long=property_data['longitude'],
        sqft=property_data['livingArea'],
        lot_sqft=property_data['lotSize'],
        year_built=property_data['yearBuilt'],
        asking_price=property_data['price'],
        rentcast_pid="NA",
        listing_type="NA",
        listed_date="NA",
        days_on_market=0
    )



from abc import ABC, abstractmethod
from typing import Type

from flows.common.contingent_status import ContingentStatus
from flows.leads_cleaner_flow.leads_task import LeadsTask
from my_asana.consts import FOLLOW_UPS_UNDER_CONTRACT_SECTION, DECLINED_LEADS_SECTION
from my_asana.users import randomly_select_user_id
from my_asana.utils import update_task, add_task_to_section


class AbstractContingentHandler(ABC):
    def __init__(self, leads_property: LeadsTask):
        self.leads_property: LeadsTask = leads_property

    def handle(self):
        from datetime import date

        # Update the due date to today
        today: str = date.today().strftime("%Y-%m-%d")
        update_task(
            task_id=self.leads_property.gid,
            data={"due_on": today}
        )

        self._handle()

    @abstractmethod
    def _handle(self):
        pass

class StillForSaleHandler(AbstractContingentHandler):
    def __init__(self, leads_property: LeadsTask):
        super().__init__(leads_property)

    def _handle(self):
        print(f'Property is still for sale: {self.leads_property.address.get_full_address()}')


class RecentlyUnderContractHandler(AbstractContingentHandler):
    def __init__(self, leads_property: LeadsTask):
        super().__init__(leads_property)

    def _handle(self):
        """
            Set the property as under contract, including:
            1. Assign the task to a random user
            2. Move the task to the Follow Ups (Under Contract) section
            3. Update the due date to today
        """
        print(f'Property has recently gone under contract: {self.leads_property.address.get_full_address()}')

        # Assign the task to a random user
        update_task(
            task_id=self.leads_property.gid,
            data={"assignee": randomly_select_user_id()}
        )

        # Move the task to the Follow Ups (Under Contract) section
        add_task_to_section(
            task_id=self.leads_property.gid,
            section_id=FOLLOW_UPS_UNDER_CONTRACT_SECTION,
        )


class RecentlySoldHandler(AbstractContingentHandler):
    def __init__(self, leads_property: LeadsTask):
        super().__init__(leads_property)

    def _handle(self):
        """
            Set the property as sold, including:
            1. Move the task to the Declined section
            2. Mark the task as completed
        """
        print(f'Property has recently been sold: {self.leads_property.address.get_full_address()}')

        # Mark the task as completed
        update_task(
            task_id=self.leads_property.gid,
            data={"completed": True}
        )

        # Move the task to the Declined section
        add_task_to_section(
            task_id=self.leads_property.gid,
            section_id=DECLINED_LEADS_SECTION,
        )


class RecentlyOffMarketHandler(AbstractContingentHandler):
    def __init__(self, leads_property: LeadsTask):
        super().__init__(leads_property)

    def _handle(self):
        print(f'Property has recently been off market: {self.leads_property.address.get_full_address()}')

        # # Add a comment
        # add_comment_to_task(
        #     task_id=self.leads_property.gid,
        #     comment=get_off_market_message()
        # )
        #
        # # Move the task to the Declined section
        # add_task_to_section(
        #     task_id=self.leads_property.gid,
        #     section_id=DECLINED_LEADS_SECTION,
        #     insert_at_the_top=True
        # )
        #
        # # Mark the task as completed
        # update_task(
        #     task_id=self.leads_property.gid,
        #     data={"completed": True}
        # )


class ContingentStatusMapper:
    """
        Maps contingent statuses to their corresponding handler classes.
    """
    @staticmethod
    def get_handler(contingent_status: ContingentStatus) -> Type[AbstractContingentHandler]:
        handlers: dict[ContingentStatus, Type[AbstractContingentHandler]] = {
            ContingentStatus.FOR_SALE: StillForSaleHandler,
            ContingentStatus.UNDER_CONTRACT: RecentlyUnderContractHandler,
            ContingentStatus.RECENTLY_SOLD: RecentlySoldHandler,
            ContingentStatus.OFF_MARKET: RecentlyOffMarketHandler
        }
        return handlers[contingent_status]


import copy

from apify.zillow_scrapers.common.client import MyApifyClient
from apify.zillow_scrapers.common.config import Actor
from common.utils.address_maniplator import update_address_with_pittsburgh_ciy
from property import PropertyAddress


def scrape_zillow_with_retry(address: PropertyAddress) -> dict:
    property_data: dict = scrape_zillow(address.get_full_address())
    if not property_data['isValid']:
        # We might have different cities in the address, e.g. Baden vs. Pittsburgh. This can happen as the task is
        # originally created using Rentcast API which will fetch the address as city=Baldwin,
        # although in Zillow it will appear with City=Pittsburgh. Hence, we will first try to scrape the data with the
        # original city, and if it fails, we will try to scrape the data with the city set to Pittsburgh.
        print(f"Property data is not valid for: {address.get_full_address()}")
        updated_address: PropertyAddress = update_address_with_pittsburgh_ciy(address)
        property_data: dict = scrape_zillow(updated_address.get_full_address())
        if not property_data['isValid']:
            raise ValueError(f"Property data is not valid for: {updated_address.get_full_address()}")
    return property_data


def scrape_zillow(address: str) -> dict:
    """
    Scrape the Zillow website using the Zillow Detail Scraper actor.
    :param address:
    :return:
    """
    # Initialize the Apify client
    apify_client: MyApifyClient = MyApifyClient(actor=Actor.ZILLOW_DETAIL_SCRAPER)

    # Prepare the Actor input
    run_input = {
        "addresses": [address]
    }

    # Fetch and print Actor results from the run's dataset (if there are any)
    items: list = apify_client.run_client(run_input=run_input)
    if len(items) > 1:
        raise ValueError(f"Expected only one item in the list, but got {len(items)} items.")

    property_data: dict = items[0]
    return property_data



import streamlit as st
import sys, os


sys.path.append(os.path.abspath("src"))

from app_flows.analyze_lead import run_single_lead_analyzer_flow
from app_flows.leads_cleaner import run_leads_cleaner_flow
from app_flows.under_contract_follow_up import run_under_contract_follow_up_flow
from app_flows.leads_generator import run_leads_generator_flow

#st.sidebar.title("📋 Flow Selector")
st.set_page_config(page_title="IHopes25To40 Tool", layout="centered")
st.title("🛠️ IHopes25To40 Tools Box 🛠️")

flow = st.sidebar.selectbox(
    "Choose a flow:",
    [
        "Single Lead Analyzer",
        "Leads Generator",
        "Leads Cleaner",
        "Under Contract Organizer"
    ]
)

if flow == "Single Lead Analyzer":
    run_single_lead_analyzer_flow()
elif flow == "Leads Generator":
    run_leads_generator_flow()
elif flow == "Leads Cleaner":
    run_leads_cleaner_flow()
elif flow == "Under Contract Organizer":
    run_under_contract_follow_up_flow()
else:
    raise ValueError("Invalid flow selected.")



from copy import copy
from typing import List

from tqdm import tqdm

from common.utils.logger import log
from flows.leads_generator_flow.flow.lead_flow import lead_flow
from flows.leads_generator_flow.flow.lead_source import LeadQuerySource, LeadsQuerySourcesHolder
from flows.leads_generator_flow.leads_filters.filters import leads_filters
from flows.leads_generator_flow.rentcast.fetch_for_sale_properties import fetch_active_for_sale_properties
from property import ForSaleProperty


def run_generate_leads_flow():
    print('Main flow - START!')
    for lead_source in LeadsQuerySourcesHolder().leads_sources:
        generate_single_source_leads(lead_source=lead_source)


def generate_single_source_leads(lead_source: LeadQuerySource) -> None:
    """
    Generate leads for a single source
    :param lead_source:
    :return:
    """
    log(msg=f'Fetching active for sale properties for {lead_source.to_string()}')
    filtered_for_sale_properties: List[ForSaleProperty] = get_active_for_sale_properties(lead_source)

    # Iterate over the properties and generate tasks
    for for_sale_property in tqdm(filtered_for_sale_properties, desc="Processing"):
        try:
            property_link: str = lead_flow(lead_property=for_sale_property)
            if property_link:
                log(msg=f'Property link: {property_link}')
            else:
                log(msg=f'No task generated for {for_sale_property.address.get_full_address()}')
        except Exception as e:
            log(msg=f'Failed to process property: {for_sale_property.address.get_full_address()} [error={e}]')


def get_active_for_sale_properties(lead_source: LeadQuerySource) -> List[ForSaleProperty]:
    """
    Get active for sale properties for a given lead source, filtered by our system filters.
    :param lead_source:
    :return:
    """
    curr_lead_source_properties: List[ForSaleProperty] = fetch_active_for_sale_properties(
        lead_source=lead_source
    )
    log(msg=f'Fetched {len(curr_lead_source_properties)} active for sale properties for {lead_source.to_string()}')

    filtered_for_sale_properties: List[ForSaleProperty] = copy(curr_lead_source_properties)

    # Filter the leads (active for sale properties)
    for leads_filter in leads_filters:
        filtered_for_sale_properties: List[ForSaleProperty] = leads_filter.apply_filter(filtered_for_sale_properties)
        log(msg=f'After applying filter {leads_filter.name}, {len(filtered_for_sale_properties)} properties left')

    return filtered_for_sale_properties


if __name__ == '__main__':
    run_generate_leads_flow()



from typing import Optional

import requests
import random
from comps_extractor.utils.coordinates import AddressCoordinates
from property import PropertyAddress


def get_coordinates(address: PropertyAddress) -> Optional[AddressCoordinates]:
    """
    Use the Nominatim API to get the latitude and longitude of an address.
    Consider add sleep time to avoid rate limiting.
    :param address: without zip code, e.g. "152 W Patty Ln, Monroeville, PA"
    :return: latitude and longitude
    """
    base_url = "https://nominatim.openstreetmap.org/search"

    headers = {
        "User-Agent": random.choice(["25to40", "IHopes", "IHopesProperties", "25to40IHopesProperties"])
    }

    params = {
        "q": address,
        "format": "json"
    }

    params = {
        "street": address.address,
        "city": address.city,
        "state": address.state,
        "country": "USA",
        "format": "json",
    }

    response = requests.get(base_url, params=params, headers=headers)
    data = response.json()

    if data:
        location = data[0]
        latitude: float = float(location["lat"])
        longtitude: float = float(location["lon"])
        print(f"Coordinates for {address}: Latitude={latitude}, Longitude={longtitude}")

        return AddressCoordinates(
            address=address.get_full_address(),
            lat=latitude,
            lon=longtitude
        )
    else:
        print(f"No coordinates found for {address}")
        return None



from dataclasses import dataclass, Field
from typing import Optional, List

from pydantic import BaseModel, field_validator




class LeadQuerySource(BaseModel):
    zipcode: Optional[str] = None
    neighborhood: Optional[str] = None
    city: str
    state: str


    def to_string(self) -> str:
        if self.zipcode:
            return f'zipcode: {self.zipcode} of {self.neighborhood}'
        if self.city:
            return self.city


class LeadsQuerySourcesHolder:
    def __init__(self):
        self.leads_sources: List[LeadQuerySource] = [
            # Inner city neighborhoods by zip code
            # LeadQuerySource(zipcode=str(BEECHVIEW), neighborhood='Beechview', city='Pittsburgh', state='PA'),
            # LeadQuerySource(zipcode=str(BROOKLINE), neighborhood='Brookline', city='Pittsburgh', state='PA'),
            # LeadQuerySource(zipcode=str(MOUNT_WASHINGTON), neighborhood='Mount Washington', city='Pittsburgh', state='PA'),
            # LeadQuerySource(zipcode=str(BRIGHTON_HEIGHTS), neighborhood='Brighton Heights', city='Pittsburgh', state='PA'),
            # LeadQuerySource(zipcode=str(GREENFIELD), neighborhood='Greenfield', city='Pittsburgh', state='PA'),
            # LeadQuerySource(zipcode=str(BLOOMFIELD), neighborhood='Bloomfield', city='Pittsburgh', state='PA'),
            # LeadQuerySource(zipcode=str(SQUIRREL_HILL), neighborhood='Squirrel Hill', city='Pittsburgh', state='PA'),

            LeadQuerySource(city='Pittsburgh', state='PA'),

            # Outside city neighborhoods by city
            LeadQuerySource(city='Monroeville', state='PA'),
            LeadQuerySource(city='Whitehall', state='PA'),
            LeadQuerySource(city='Baldwin', state='PA'),
            LeadQuerySource(city='Carnegie', state='PA'),
            LeadQuerySource(city='Scott Township', state='PA'),
            LeadQuerySource(city='Bethel Park', state='PA'),
            LeadQuerySource(city='Avalon', state='PA'),
            LeadQuerySource(city='Bellevue', state='PA'),
        ]

    def get_source_by_city(self, city: str) -> Optional[LeadQuerySource]:
        """
        Get a lead source by city
        :param city:
        :return: A LeadQuerySource object if found, None otherwise.
        """
        for lead_source in self.leads_sources:
            if lead_source.city == city:
                return lead_source
        return None

    def get_all_cities(self) -> List[str]:
        """
        Get all cities from the lead sources
        :return: A list of cities
        """
        return [lead_source.city for lead_source in self.leads_sources]



import streamlit as st

from flows.under_contract_follow_up_flow.flow import run_under_contract_leads_flow


def run_under_contract_follow_up_flow():
    st.title("📞 Under Contract Follow-Up")
    st.write("Track, manage and organize our current under contract leads.")

    if st.button("Organize Under Contract Properties"):
        with st.spinner("Running..."):
            try:
                st.info(f"Under contract flow has been triggered.")
                run_under_contract_leads_flow()
                st.success(f"Under contract flow has been completed.")

            except Exception as e:
                st.error(f"Error: {e}")



from typing import List

from flows.leads_generator_flow.leads_filters.abstract_filter import LeadsFilter
from property import ForSaleProperty


class SqftFilter(LeadsFilter):
    """
    Filter properties by number of bedrooms
    """

    def __init__(self, min_sqft: int):
        super().__init__('Sqft Filter')
        self.min_sqft: int = min_sqft

    def apply_filter(self, leads: List[ForSaleProperty]) -> List[ForSaleProperty]:
        return [lead for lead in leads if not lead.sqft or self.min_sqft <= lead.sqft]


SQFT_FILTER: SqftFilter = SqftFilter(
    min_sqft=800
)



from typing import List

from flows.leads_generator_flow.leads_filters.abstract_filter import LeadsFilter
from flows.leads_generator_flow.leads_filters.ap_filter import ASKING_PRICE_FILTER
from flows.leads_generator_flow.leads_filters.bedrooms_filter import BEDROOMS_FILTER
from flows.leads_generator_flow.leads_filters.internal_neighborhoods_filter import IN_CITY_NEIGHBORHOODS_FILTER
from flows.leads_generator_flow.leads_filters.property_sqft_filter import SQFT_FILTER

leads_filters: List[LeadsFilter] = [
    IN_CITY_NEIGHBORHOODS_FILTER, # Has to be the first filter
    ASKING_PRICE_FILTER,
    BEDROOMS_FILTER,
    SQFT_FILTER
]



from typing import List

from googleapiclient.discovery import Resource

from google_drive.authenticator import get_google_services
from google_drive.consts import ARV_TEMPLATE_FILE_ID
from google_drive.spreadsheet_filler import write_comps_to_sheet, write_lead_property_info_to_sheet
from property import SoldProperty, ForSaleProperty


class CompsSpreadsheet:
    def __init__(self, file_id: str, file_path: str):
        self.file_id = file_id
        self.file_path = file_path

    def __str__(self):
        return f"File ID: {self.file_id}, File path: {self.file_path}"


def copy_google_sheet(service: Resource, source_file_id: str, new_file_name: str) -> CompsSpreadsheet:
    """
    Function to copy the file and rename it
    :param service:
    :param source_file_id:
    :param new_file_name:
    :return:
    """
    # Create a copy of the file
    copied_file = service.files().copy(
        fileId=source_file_id,
        body={'name': new_file_name}
    ).execute()

    print(f"File copied and renamed to: {copied_file['name']}")
    file_path: str = f'https://docs.google.com/spreadsheets/d/{copied_file["id"]}'
    return CompsSpreadsheet(
        file_id=copied_file['id'],
        file_path=file_path
    )


def save_as_template_arv(for_sale_property: ForSaleProperty, comps: List[SoldProperty]) -> str:
    """
    Save the ARV template with the lead property and comps data.
    :param for_sale_property:
    :param comps:
    :return:
    """
    drive_service: Resource = get_google_services().drive
    sheets_service: Resource = get_google_services().sheets

    # Copy the file and rename it
    comps_doc: CompsSpreadsheet = copy_google_sheet(
        service=drive_service,
        source_file_id=ARV_TEMPLATE_FILE_ID,
        new_file_name=for_sale_property.address.get_full_address()
    )

    # Write the lead property info to the sheet
    write_lead_property_info_to_sheet(
        service=sheets_service,
        sheet_id=comps_doc.file_id,
        for_sale_property=for_sale_property
    )

    # Write comps to the sheet
    write_comps_to_sheet(
        service=sheets_service,
        sheet_id=comps_doc.file_id,
        comps=comps
    )

    return comps_doc.file_path



from typing import List

from flows.leads_generator_flow.leads_filters.abstract_filter import LeadsFilter
from property import ForSaleProperty


class BedroomsFilter(LeadsFilter):
    """
    Filter properties by number of bedrooms
    """

    def __init__(self, min_bedrooms: int, max_bedrooms: int):
        super().__init__('Bedrooms Filter')
        self.min_bedrooms: int = min_bedrooms
        self.max_bedrooms: int = max_bedrooms

    def apply_filter(self, leads: List[ForSaleProperty]) -> List[ForSaleProperty]:
        return [lead for lead in leads if self.min_bedrooms <= lead.bedrooms <= self.max_bedrooms]


BEDROOMS_FILTER: BedroomsFilter = BedroomsFilter(
    min_bedrooms=2,
    max_bedrooms=4
)



#APIFY_API_TOKEN = 'apify_api_RiHYgad2cEbJVzHCMEcECjMaGCmZ4145Zemw'
#APIFY_API_TOKEN = 'apify_api_3BOnahpympcgHIFZNAwEa5N4dhlJ3i2GCrzu'
from enum import Enum

APIFY_API_TOKEN = 'apify_api_FMtMdLbfYGOtbmkX4asDkztQsf4gJX1iVX2R'


class Actor(Enum):
    ZILLOW_SEARCH_SCRAPER = 'maxcopell/zillow-scraper'
    ZILLOW_DETAIL_SCRAPER = 'maxcopell/zillow-detail-scraper'



from abc import ABC, abstractmethod
from typing import Type

from flows.common.contingent_status import ContingentStatus
from flows.under_contract_follow_up_flow.under_contract_task import UnderContractTask
from my_asana.comments_helper import get_up_for_sale_again_message, add_comment_to_task, \
    add_followers_to_task, get_property_sold_message, get_off_market_message
from my_asana.consts import LEADS_PROPERTIES_SECTION, DECLINED_LEADS_SECTION
from my_asana.users import randomly_select_user_id
from my_asana.utils import add_task_to_section, update_task



class AbstractContingentHandler(ABC):
    def __init__(self, under_contract_property: UnderContractTask):
        self.under_contract_property: UnderContractTask = under_contract_property

    @abstractmethod
    def handle(self):
        pass


class StillUnderContractHandler(AbstractContingentHandler):
    def __init__(self, under_contract_property: UnderContractTask):
        super().__init__(under_contract_property)

    def handle(self):
        print(f'Property is still under contract: {self.under_contract_property.address.get_full_address()}')


class RecentlySoldHandler(AbstractContingentHandler):
    def __init__(self, under_contract_property: UnderContractTask):
        super().__init__(under_contract_property)

    def handle(self):
        """
        Set the property as recently sold, including:
        1. Add a comment to the task (notify the property is sold)
        2. Mark the task as completed
        3. Add the task to the Declined section
        :param under_contract_property:
        :return:
        """
        print(f'Property is recently sold: {self.under_contract_property.address.get_full_address()}')

        # Add a comment
        add_comment_to_task(
            task_id=self.under_contract_property.gid,
            comment=get_property_sold_message()
        )

        # Mark the task as completed
        update_task(
            task_id=self.under_contract_property.gid,
            data={"completed": True}
        )

        # Add the task to the Declined section
        add_task_to_section(
            task_id=self.under_contract_property.gid,
            section_id=DECLINED_LEADS_SECTION,
        )


class ForSaleHandler(AbstractContingentHandler):
    def __init__(self, under_contract_property: UnderContractTask):
        super().__init__(under_contract_property)

    def handle(self):
        """
        Set the property as for sale again, including:
        1. Add followers (to make sure the tagging will directly notify them)
        2. Add a comment to the task (notify the property is up for sale again)
        3. Assign the task to a random user
        4. Add the task to the Leads section
        :param under_contract_property:
        :return:
            """
        print(f'Property is up for sale again: {self.under_contract_property.address.get_full_address()}')

        # Add followers
        add_followers_to_task(
            task_id=self.under_contract_property.gid
        )

        # Add a comment
        add_comment_to_task(
            task_id=self.under_contract_property.gid,
            comment=get_up_for_sale_again_message()
        )

        # Assign the task to a random user
        update_task(
            task_id=self.under_contract_property.gid,
            data={"assignee": randomly_select_user_id()}
        )

        # Add the task to the Leads section
        add_task_to_section(
            task_id=self.under_contract_property.gid,
            section_id=LEADS_PROPERTIES_SECTION,
            insert_at_the_top=True
        )

class OffMarketHandler(AbstractContingentHandler):
    def __init__(self, under_contract_property: UnderContractTask):
        super().__init__(under_contract_property)

    def handle(self):
        """
            Set the property as off market, including:
            1. Add a comment to the task (notify the property is off market)
            2. Mark the task as completed
            3. Add the task to the Declined section
            :param under_contract_property:
            :return:
            """
        print(f'Property is off market: {self.under_contract_property.address.get_full_address()}')

        # Add a comment
        add_comment_to_task(
            task_id=self.under_contract_property.gid,
            comment=get_off_market_message()
        )

        # Mark the task as completed
        update_task(
            task_id=self.under_contract_property.gid,
            data={"completed": True}
        )

        # Add the task to the Declined section
        add_task_to_section(
            task_id=self.under_contract_property.gid,
            section_id=DECLINED_LEADS_SECTION,
        )

class ContingentStatusMapper:
    """
        Maps contingent statuses to their corresponding handler classes.
    """
    @staticmethod
    def get_handler(contingent_status: ContingentStatus) -> Type[AbstractContingentHandler]:
        handlers: dict[ContingentStatus, Type[AbstractContingentHandler]] = {
            ContingentStatus.UNDER_CONTRACT: StillUnderContractHandler,
            ContingentStatus.FOR_SALE: ForSaleHandler,
            ContingentStatus.RECENTLY_SOLD: RecentlySoldHandler,
            ContingentStatus.OFF_MARKET: OffMarketHandler
        }
        return handlers[contingent_status]


from typing import Optional

import streamlit as st

from flows.leads_generator_flow.flow.generate_leads import generate_single_source_leads, run_generate_leads_flow
from flows.leads_generator_flow.flow.lead_source import LeadsQuerySourcesHolder, LeadQuerySource


def run_leads_generator_flow():
    st.header("🚀 Leads Generator Flow 🚀")
    st.write("Select a source to generate new leads from:")

    leads_sources_holder: LeadsQuerySourcesHolder = LeadsQuerySourcesHolder()
    options = leads_sources_holder.get_all_cities()
    options.append("All")  # Add a custom option for manual input

    selected_option = st.selectbox("Lead Source", options)

    confirm_all = False
    if selected_option == "All":
        st.warning("⚠️ You selected **All**. This will trigger all lead generation sources and may take time "
                   "and cost some money.")
        confirm_all = st.checkbox("I understand, proceed with All sources")

    if st.button("Generate Leads"):
        if selected_option == "All":
            if not confirm_all:
                st.warning("Please confirm you want to proceed with all sources.")
                return

            st.info("Running leads generation from: **All Sources**...")
            try:
                run_generate_leads_flow()
                st.success("✅ Successfully generated leads from all sources!")
            except Exception as e:
                st.error(f"❌ Failed to generate leads: {e}")
        else:
            st.info(f"Running leads generation from: **{selected_option}**...")
            try:
                lead_source: Optional[LeadQuerySource] = leads_sources_holder.get_source_by_city(city=selected_option)
                if not lead_source:
                    st.warning(f"No lead source found for city: {selected_option}")
                    return
                generate_single_source_leads(lead_source=lead_source)
                st.success(f"✅ Successfully generated leads from {selected_option}!")
            except Exception as e:
                st.error(f"❌ Failed to generate leads: {e}")



import pywhatkit as kit

TWENTY_TO_FORTHY_GROUP_ID: str = 'CMdxIFPaTzq0bOFU8BxaJj'

#kit.sendwhatmsg_instantly("+972545336285", "Hello!")

def notify_completion(new_leads_count: int) -> None:
    kit.sendwhatmsg_to_group_instantly(
        group_id=TWENTY_TO_FORTHY_GROUP_ID,
        message=f'[Leads Flow System Bot] I have just generated {new_leads_count} new leads for you!'
    )
    print(f'Whatsapp message sent to group {TWENTY_TO_FORTHY_GROUP_ID} with the new leads count')

notify_completion(80)


from enum import Enum
from property import PropertyAddress


class UnderContractTask:
    def __init__(self, address: str, gid: str):
        self.address: PropertyAddress = PropertyAddress.from_full_address(address)
        self.gid: str = gid






from typing import List

from flows.leads_generator_flow.leads_filters.abstract_filter import LeadsFilter
from property import ForSaleProperty


class AskingPriceFilter(LeadsFilter):
    """
    White list of neighborhoods
    """

    def __init__(self, min_price: int, max_price: int):
        super().__init__('Asking Price Filter')
        self.min_price: int = min_price
        self.max_price: int = max_price

    def apply_filter(self, leads: List[ForSaleProperty]) -> List[ForSaleProperty]:
        return [lead for lead in leads if self.min_price <= lead.asking_price <= self.max_price]


ASKING_PRICE_FILTER: AskingPriceFilter = AskingPriceFilter(
    min_price=80000,  # 80K
    max_price=210000  # 210K
)



# setup.py

from setuptools import setup, find_packages

setup(
    name="IHopesProperties",
    version="1.1.1",
    packages=find_packages(),
    install_requires=[],
    author="Ori Peri",
    author_email="25to40@gmail.com",
    description="Leads flow of IHopesProperties",
)



from typing import Optional

from zillow.property_getter import construct_zillow_basic_property_url


class PropertyAddress:
    def __init__(self, address: str, city: str, state: str, zip_code: str):
        self.address: str = address
        self.city: str = city
        self.state: str = state
        self.zip_code: str = zip_code

    def get_full_address(self):
        return f"{self.address}, {self.city}, {self.state} {self.zip_code}"

    @staticmethod
    def from_full_address(full_address: str):
        """
        Address can either be '439 William St, Mount, Oliver, PA 15210' or '439 William St, PA 15210'.
        :param full_address:
        :return:
        """
        address_parts = full_address.split(',')
        if len(address_parts) == 4:
            address = address_parts[0].strip()
            city = f'{address_parts[1].strip()}, {address_parts[2].strip()}'
            state_zip = address_parts[3].strip().split()
            state = state_zip[0]
            zip_code = state_zip[1]
        elif len(address_parts) == 3:
            address = address_parts[0].strip()
            city = address_parts[1].strip()
            state_zip = address_parts[2].strip().split()
            state = state_zip[0]
            zip_code = state_zip[1]
        else:
            raise ValueError(f"Invalid address format: {full_address}")
        return PropertyAddress(address, city, state, zip_code)


class ListingAgent:
    def __init__(self, name: str, phone: str, email: str, website: Optional[str] = None):
        self.name: str = name
        self.phone: str = phone
        self.email: str = email
        self.website: Optional[str] = website



class Property:
    def __init__(self, address: PropertyAddress, beds: float, baths: float, sqft: int, lot_sqft: float,
                 lat: float, long: float, year_built: Optional[int] = None, zillow_link: Optional[str] = None):
        self.address: PropertyAddress = address
        self.bedrooms: float = beds
        self.bathrooms: float = baths
        self.sqft: int = sqft
        self.lot_sqft: float = lot_sqft
        self.lat: float = lat
        self.long: float = long
        self.year_built: Optional[int] = year_built
        self.zillow_link: str = zillow_link if zillow_link \
            else construct_zillow_basic_property_url(self.address.get_full_address())

    def get_min_sqft(self):
        """
        Consider up to 20% less than sqft
        :return:
        """
        return self.sqft * 0.8

    def get_max_sqft(self):
        """
        Consider up to 20% more than sqft
        :return:
        """
        return self.sqft * 1.2

    @staticmethod
    def calculate_distance(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
        """
        Calculate the great-circle distance between two points on the Earth's surface.

        Parameters:
        - lat1, lon1: Latitude and Longitude of point 1 in decimal degrees.
        - lat2, lon2: Latitude and Longitude of point 2 in decimal degrees.

        Returns:
        - Distance in miles.
        """
        # Earth radius in miles
        R = 3958.8

        # Convert latitude and longitude from degrees to radians
        import math
        lat1, lon1, lat2, lon2 = map(math.radians, [lat1, lon1, lat2, lon2])

        # Differences in latitude and longitude
        dlat = lat2 - lat1
        dlon = lon2 - lon1

        # Haversine formula
        a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

        # Distance in miles
        distance = R * c
        return distance


class ForSaleProperty(Property):
    def __init__(self, rentcast_pid: str, address: PropertyAddress, beds: float, baths: float,
                 sqft: int, lot_sqft: float, lat: float, long: float, year_built: int, asking_price: int,
                 listing_type: str, listed_date: str, days_on_market: int):
        super().__init__(
            address=address,
            beds=beds,
            baths=baths,
            sqft=sqft,
            lot_sqft=lot_sqft,
            year_built=year_built,
            lat=lat,
            long=long,
        )
        self.rentcast_pid: str = rentcast_pid
        self.asking_price: int = asking_price
        self.listing_type: str = listing_type
        self.listed_date: str = listed_date
        self.days_on_market: int = days_on_market

    def __eq__(self, other):
        return self.rentcast_pid == other.rentcast_pid

    def __hash__(self):
        return hash(self.rentcast_pid)


class SoldProperty(Property):
    def __init__(self, zillow_pid: str, address: PropertyAddress, beds: float, baths: float,
                 sqft: int, lot_sqft: float, lat: float, long: float, dist_from_lead: float, zillow_link: str,
                 sold_price: int, sold_date: str):
        super().__init__(
            address=address,
            beds=beds,
            baths=baths,
            sqft=sqft,
            lot_sqft=lot_sqft,
            lat=lat,
            long=long,
            zillow_link=zillow_link
        )
        self.dist_from_lead: float = dist_from_lead
        self.zillow_pid: str = zillow_pid
        self.sold_price: int = sold_price
        self.sold_date: str = sold_date



from typing import List, Type

from apify.zillow_scrapers.zillow_detail_scraper.contingency_status_extractor import get_contingency_status_by_address
from flows.common.contingent_status import ContingentStatus
from flows.leads_cleaner_flow.contingent_status_handlers import ContingentStatusMapper, AbstractContingentHandler
from flows.leads_cleaner_flow.get_leads_properties import get_leads_properties
from flows.leads_cleaner_flow.leads_task import LeadsTask


def run_leads_cleaner():
    """
    Run the leads cleaner flow, will update all the leads that are not active anymore.
    """
    print('Running leads cleaner flow')
    leads_properties: List[LeadsTask] = get_leads_properties()

    for leads_property in leads_properties:
        property_address: str = leads_property.address.get_full_address()
        print(f'\nLeads property: {property_address}')

        status: ContingentStatus = get_contingency_status_by_address(property_address=leads_property.address)
        print(f'For property {property_address}, status is {status.value}')

        handler_class: Type[AbstractContingentHandler] = ContingentStatusMapper().get_handler(status)
        handler_instance: AbstractContingentHandler = handler_class(leads_property)
        handler_instance.handle()
        print(f'Finished handling property: {property_address}')


if __name__ == '__main__':
    run_leads_cleaner()


from asana import CustomFieldsApi, TagsApi, UsersApi

from my_asana.authenticate import get_project_api, get_custom_fields_api, get_workspace_api, get_tags_api, get_users_api
#
# users_api: UsersApi = get_users_api()
# list(users_api.get_users(opts={'workspace':WORKSPACE_ID}))
#
#
#
# projects_api = get_project_api()
# projects_api.get_project(REAL_ESTATE_LEADS_PROJECT_ID, {"opt_fields": "custom_fields"})
#
# tags_api: TagsApi = get_tags_api()
# #list(tags_api.get_tags(opts={}))
#
# list(tags_api.get_tags(opts={'workspace':WORKSPACE_ID}))
# print(tags_api.get_tags_for_workspace(LEADS_PROJECT_ID))


#workspace_api = get_workspace_api()

#workspaces = workspace_api.get_workspaces()

# for workspace in workspaces.data:
#     print(f"Workspace ID: {workspace.gid} - Name: {workspace.name}")


# projects_api = get_project_api()
# project_id = LEADS_PROJECT_ID
#
# project = projects_api.get_project(project_id, {"opt_fields": "custom_fields"})
# for field in project.data["custom_fields"]:
#     print(field["gid"], field["name"])
#
# custom_fields_api = get_custom_fields_api()
# priority_field = custom_fields_api.get_custom_field(priority_field_id)
# for option in priority_field.enum_options:
#     print(option.gid, option.name)



import os
import pickle
import sys
from functools import lru_cache
from pathlib import Path
from typing import Tuple, Optional
from functools import lru_cache

from google_auth_oauthlib.flow import InstalledAppFlow, Flow
from googleapiclient.discovery import build, Resource
import streamlit as st

from common.utils.logger import log


def is_streamlit_runtime() -> bool:
    """
    Check if the script is running in a Streamlit runtime environment.
    :return:
    """
    return any("streamlit" in arg for arg in sys.argv)


CACHE_DECORATOR = st.cache_resource if is_streamlit_runtime() else lru_cache(maxsize=1)


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class GoogleServices(metaclass=SingletonMeta):
    def __init__(self):
        self._creds = None
        self.drive: Optional[Resource] = None
        self.sheets: Optional[Resource] = None

        self.scopes = [
            'https://www.googleapis.com/auth/drive',
            'https://www.googleapis.com/auth/spreadsheets'
        ]

        self.authenticate()

    def authenticate(self) -> None:
        # Authenticate and create the service
        import streamlit as st
        print("Authenticating with Google Drive and Sheets API...")
        creds = None
        # The file token.pickle stores the user's access and refresh tokens.
        parent_path: Path = Path(os.path.abspath(__file__)).parent
        token_file_path: Path = Path(parent_path, 'token.pickle')

        os_path_exists = os.path.exists(token_file_path)
        if os_path_exists:
            print("Loading credentials from token.pickle")
            with open(token_file_path, 'rb') as token:
                # from google.oauth2.credentials import Credentials
                # creds2 = Credentials.from_authorized_user_file('token.json', scopes=['your-scopes'])
                creds = pickle.load(token)
                print("Credentials loaded successfully.")
        else:
            print("token.pickle file not found. Starting authentication process.")

        if not creds or not creds.valid:
            try:
                from google.auth.transport.requests import Request
                creds.refresh(Request())
                log(msg="Authenticated successfully to Google Drive and Sheets API.")
            except Exception as e:
                st.write("Canot do refresh ")
            if not creds or not creds.valid:
                print("Credentials not found or invalid. Starting authentication process.")
                st.write("Credentials not found or invalid. Starting authentication process.")
                if is_streamlit_runtime():
                    print("Running in Streamlit runtime. Using OAuth 2.0 authorization flow.")
                    # Define your redirect URI based on Streamlit Cloud
                    streamlit_base_url = "https://leads-generator-ihopes25to40.streamlit.app"
                    redirect_uri = f"{streamlit_base_url}/"

                    flow = Flow.from_client_secrets_file(
                        client_secrets_file=str(parent_path / 'credentials.json'),
                        scopes=self.scopes,
                        redirect_uri=redirect_uri
                    )

                    auth_url, _ = flow.authorization_url(
                        prompt='consent',
                        access_type='offline',
                        include_granted_scopes='true'
                    )

                    st.markdown("### 🔐 Google Authentication Required")
                    st.markdown(f"[Click here here here here here to authenticate]({auth_url})")
                    code = st.text_input("Paste the authorization code from the URL after granting access:")

                    if code:
                        try:
                            flow.fetch_token(code=code)
                            creds = flow.credentials
                            with open(token_file_path, 'wb') as token:
                                pickle.dump(creds, token)
                            st.success("✅ Authentication successful!")
                        except Exception as e:
                            st.error(f"❌ Authentication failed: {e}")
                    else:
                        st.stop()
                else:
                    print("Running outside Streamlit runtime. Using local server for authentication.")
                    flow = InstalledAppFlow.from_client_secrets_file(
                        client_secrets_file=f'{parent_path}/credentials.json',
                        scopes=self.scopes,
                    )
                    creds = flow.run_local_server(port=8080)
                    with open(token_file_path, 'wb') as token:
                        pickle.dump(creds, token)

        # Build the Drive API service
        self.drive: Resource = build('drive', 'v3', credentials=creds)
        self.sheets: Resource = build('sheets', 'v4', credentials=creds)


def get_google_services() -> GoogleServices:
    return GoogleServices()



from typing import List

from googleapiclient.discovery import Resource

from property import SoldProperty, ForSaleProperty


def write_lead_property_info_to_sheet(service: Resource, sheet_id: str, for_sale_property: ForSaleProperty) -> None:
    """
    Writes the property information to the specified Google Sheet.
    :param service:
    :param sheet_id:
    :param for_sale_property:
    :return:
    """
    sheets = service.spreadsheets()

    # Fetch the sheet data to find where "Property Info" is located
    sheet_data = sheets.values().get(spreadsheetId=sheet_id, range='ARV & Comps').execute()
    values = sheet_data.get('values', [])

    # Prepare data for insertion
    property_data = [
        [
            for_sale_property.zillow_link,  # Property Link
            for_sale_property.address.get_full_address(),
            for_sale_property.listed_date,
            f"{for_sale_property.bedrooms}/{for_sale_property.bathrooms}",
            for_sale_property.sqft,
            for_sale_property.lot_sqft,
            for_sale_property.year_built,
            "",  # Empty column currently
            "",  # Condition (optional, if applicable)
            "",  # Special Features (optional, if applicable)
            for_sale_property.asking_price
        ]
    ]

    # Write the property data to the sheet
    range_to_update = f'ARV & Comps!A2:K3'
    sheets.values().update(
        spreadsheetId=sheet_id,
        range=range_to_update,
        valueInputOption='RAW',
        body={'values': property_data}
    ).execute()


def write_comps_to_sheet(service: Resource, sheet_id: str, comps: List[SoldProperty]) -> None:
    """
    Writes the comps data to the specified Google Sheet.
    :param service: Google Sheets API resource.
    :param sheet_id: ID of the Google Sheet.
    :param comps: List of Property objects to write.
    """
    sheets = service.spreadsheets()

    # Fetch the sheet data to find where "Comps" is located
    sheet_data = sheets.values().get(spreadsheetId=sheet_id, range='ARV & Comps').execute()
    values = sheet_data.get('values', [])

    # Locate the "Comps" row
    comps_row_index = next(
        (index for index, row in enumerate(values) if "Comps" in row), None
    )
    if comps_row_index is None:
        raise ValueError("Comps section not found in the template.")

    # Starting row for the comps data
    start_row = comps_row_index + 3  # Three rows below "Comps" for headers and since it's 0-based index

    # Prepare data for insertion
    comps_data = [
        [
            comp.zillow_link,  # Property Link
            comp.address.get_full_address(),
            comp.sold_date,  # Listing Sold Date
            f"{comp.bedrooms}/{comp.bathrooms}",
            comp.sqft,
            comp.lot_sqft,  # Lot Sqf
            comp.year_built,  # Year Built (optional, if applicable)
            comp.dist_from_lead,  # distance from lead property
            "",  # Condition (optional, if applicable)
            "",  # Special Features (optional, if applicable)
            comp.sold_price
        ]
        for comp in comps
    ]

    # Write the comps data to the sheet
    range_to_update = f'ARV & Comps!A{start_row}:K{start_row + len(comps_data) - 1}'
    sheets.values().update(
        spreadsheetId=sheet_id,
        range=range_to_update,
        valueInputOption='RAW',
        body={'values': comps_data}
    ).execute()



from typing import List

from flows.leads_generator_flow.leads_filters.abstract_filter import LeadsFilter
from property import ForSaleProperty


class OutsideCityNeighborhoodFilter(LeadsFilter):
    """
    White list of neighborhoods
    """

    def __init__(self, white_list_neighborhoods: List[str]):
        super().__init__('Neighborhoods Filter')
        self.white_list_neighborhoods: List[str] = white_list_neighborhoods

    def apply_filter(self, leads: List[ForSaleProperty]) -> List[ForSaleProperty]:
        return [lead for lead in leads if lead.address.city in self.white_list_neighborhoods]


OUTSIDE_NEIGHBORHOODS_FILTER: OutsideCityNeighborhoodFilter = OutsideCityNeighborhoodFilter(
    white_list_neighborhoods=[
        "Pittsburgh", "Monroeville", "Dormont", "Castle Shannon", "Whitehall", "Brentwood",
        "Baldwin", "Mt. Lebanon", "Carnegie", "Scott Township", "MT Washington", "Bethel Park",
        "Greentree", "Crafton", "Avalon", "Bellevue", "Swissvale"
    ]
)



from typing import Optional

from apify_client import ApifyClient

from apify.zillow_scrapers.common.config import APIFY_API_TOKEN, Actor


def get_apify_client() -> ApifyClient:
    client = ApifyClient(APIFY_API_TOKEN)
    return client



class MyApifyClient:

    def __init__(self, actor: Actor):
        self.client = ApifyClient(APIFY_API_TOKEN)
        self.actor: Actor = actor

    def run_client(self, run_input: dict, test_dataset_id: Optional[str] = None) -> Optional[list]:
        """
        Run one of the Apify actors to scrape zillow data.
        :param test_dataset_id: If running in test mode, use this dataset ID.
        :param run_input:
        :return:
        """
        if test_dataset_id:
            dataset_id: str = test_dataset_id
        else:
            run = self.client.actor(self.actor.value).call(run_input=run_input)
            dataset_id: str = run["defaultDatasetId"]

            if 'scraped 0 items' in run['statusMessage']:
                print("No data found in the search results.")
                return None

        print("Check your data here: https://console.apify.com/storage/datasets/" + dataset_id)

        items: list = self.client.dataset(dataset_id).list_items().items
        if not items:
            print("No items found in the dataset.")
            return None

        return items



from asana import StoriesApi

from common.utils.address_maniplator import update_address_with_pittsburgh_ciy
from my_asana.authenticate import get_tasks_api, get_stories_api
from my_asana.consts import LEADS_PROJECT_ID, LEADS_PROPERTIES_SECTION, PROPERTY_TASK_TEMPLATE_ID
from property import Property, PropertyAddress


def is_task_exists(property_address: str) -> bool:
    tasks_api = get_tasks_api()
    tasks = tasks_api.get_tasks_for_project(LEADS_PROJECT_ID, opts={})
    updated_property_address: str = update_address_with_pittsburgh_ciy(
        address=PropertyAddress.from_full_address(property_address)).get_full_address()
    for task in tasks:
        if task['name'] == property_address or task['name'] == updated_property_address:
            print(f"Task already exists for: {property_address}")
            return True
    return False


def add_task_to_section(task_id: str, section_id: str, project_id: str = LEADS_PROJECT_ID,
                        insert_at_the_top: bool = False):
    """
    Add a task to a specific section in a project
    :param task_id: task ID to add
    :param section_id: section ID to add the task to
    :param project_id: project ID to add the task to
    :param insert_at_the_top: boolean to insert the task at the top of the section
    :return:
    """
    if project_id != LEADS_PROJECT_ID and insert_at_the_top:
        raise RuntimeError("insert_at_the_top can only be used with the LEADS_PROJECT_ID project."
                           " When using a different project, you should set insert_at_the_top to False.")

    tasks_api = get_tasks_api()

    data = {
        "data": {
            "project": project_id,
        }
    }
    if insert_at_the_top:
        data['data']['insert_after'] = PROPERTY_TASK_TEMPLATE_ID
    else:
        data['data']['section'] = section_id

    # Add the task to the project and section using the 'add_project_for_task' method
    tasks_api.add_project_for_task(task_gid=task_id, body=data)
    print(f'Added task {task_id} to section {section_id}')


def update_task(task_id: str, data: dict):
    """
    Update a task using the Asana API
    :param task_id: task ID to update
    :param data: data to update the task with
    :return:
    """
    tasks_api = get_tasks_api()
    tasks_api.update_task(
        task_gid=task_id,
        body={"data": data},
        opts={}
    )
    _costume_update_task_output(task_id, data)

def _costume_update_task_output(task_id: str, data: dict):
    """
    Custom output for the update_task function
    :param task_id: task ID to update
    :param data: data to update the task with
    :return:
    """
    str_output: str = f'Updated task {task_id}'

    if 'assignee' in data:
        str_output += f' with assignee {data["assignee"]}'
    if 'due_on' in data:
        str_output += f' with due date {data["due_on"]}'
    if 'completed' in data:
        str_output += f' as completed'

    print(str_output)